/* This program is free software. It comes without any warranty, to
 * the extent permitted by applicable law. You can redistribute it
 * and/or modify it under the terms of the Do What The Fuck You Want
 * To Public License, Version 2, as published by Sam Hocevar. See
 * http://sam.zoy.org/wtfpl/COPYING for more details. */ 

#include <stdio.h>
#include <stdlib.h>
#include <map> 
#include <list> 
#include <set> 
#include <string> 
#include <string.h> 
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <openssl/md5.h>
#include <dirent.h> 
#include <fcntl.h>
#include <assert.h>

/*
  todo:
  - discard useless sim_dir ? sort by % ?
  - interactive f et d 
  - virer O_DIRECT ?
*/

void printreadme () {
  printf(
	 "Searches and reports duplicate files and directories:\n"
	 "- two files are identicals if the MD5 hash is the same (in id_files.txt)\n"
	 "- reports two dirs as identical if they have the same set of hash files (ignore filenames, empty files and dirs) (in id_dirs.txt)\n"
	 "- if the two diretories are realy equal (same filenames, empty files ans dirs), it reports it with \"EXACT\" flag\n"
	 "- try to find \"similar\" dirs: dir A is a subset of dir B, or A inter B is big\n"
	 "- reports are sorted by the saved space size (reverse order)\n"
	 "\n"
	 "- do not consider or follow symlinks, or block/char files\n"
	 "- do not report hardlinked files as duplicate\n"
	 "- do not make a byte to byte comparaison (useless...)\n"
	 "\n"
	 "It tries to optimise the hash computation :\n" 
	 "- on a first pass, it compute only hash of first and last 4k data of larges files\n"
	 "- sort files by inode number before hashing\n"
	 "\n"
	 "Also:\n" 
	 "- report hard links in hl.txt\n"
	 "- report directories with a lot of files in lof_dirs.txt\n"
	 "- reports different files with same begining in sim_files.txt (useful for mp3 and containers with tags)\n"   
	 "\n"
	 "Known bugs :\n"
	 "- may use a lot of memory...\n"
	 "- EXACT dirs may be badly reported if there is more than two identical dirs\n"
	 );
}

void printusage () {
  printf("fdupdirs V 0.1\n" 
	 "Searches and reports duplicate files and directories (-H for more infos)\n"
	 "Usage: fdupdirs [options] path [ path [...] ]\n"
	 "\n"
	 "Options:\n"
	 " -h : print this help\n"
	 " -H : print more infos on this program\n"
	 " -n : do not perform a md5sum on whole files\n"
	 " -t : sort dirs by size (instead of saved space)\n"
	 " -r x : report duplicates of size at least x bytes (default: 2^20)\n"
	 " -s x : perform an prehash for every file of size at least x bytes (default: 2^20)\n"
	 " -b : use only the first 4k block to perform prehash\n"
	 );
}

using namespace std;

typedef long long ll_t;

char *human(ll_t x,char *bf=NULL,int s=30) 
{
  static char bfhuman[30];
  if(bf==NULL) bf=bfhuman;
  if(x<1000) snprintf(bf,s,"%Ldb",x);
  else if(x<1000LL*1024) snprintf(bf,s,"%.1fk",x/1024.);
  else if(x<1000LL*1024*1204) snprintf(bf,s,"%.1fM",x/1024./1024.);
  else if(x<1000LL*1024*1204*1024) snprintf(bf,s,"%.1fG",x/1024./1024./1024.);
  else snprintf(bf,s,"%.1fT",x/1024./1024./1024./1024.);
  return bf;
}

bool md5=true;
bool noend=false;
ll_t systematic=1024LL*1024;
ll_t minreportsize=1024LL*1024;
bool sortbysize=false;

struct hash_t {
  unsigned char hash[16];
  bool operator<(const hash_t &h) const {
    for(int i=0;i<16;i++) {
      if(hash[i]<h.hash[i]) return 1;
      if(hash[i]>h.hash[i]) return 0;
    }
    return 0;
  }
  bool operator==(const hash_t &h) const {
    for(int i=0;i<16;i++) {
      if(hash[i]!=h.hash[i]) return 0;
    }
    return 1;
  }
  bool operator!=(const hash_t &h) const {
    for(int i=0;i<16;i++) {
      if(hash[i]!=h.hash[i]) return 1;
    }
    return 0;
  }
  void setx(int x) {
    memset(hash,x,16);
  }
  bool isx(int x) const {
    for(int i=0;i<16;i++) {
      if(hash[i]!=x) return 0;
    }
    return 1;
  }
  bool isx() const {
    for(int i=0;i<3;i++) {
      if(isx(i)) return 1;
    }
    return 0;
  }
  bool issol() const {
    return isx(0);
  }
  bool isunk() const {
    return isx(1);
  }
  bool isskip() const {
    return isx(2);
  }

  bool setskip() {
    setx(2);
  }
  bool setsol() {
     setx(0);
  }
  bool setunk() {
    setx(1);
  }

  hash_t(int i=1) {
    setunk();
  }
  hash_t(const hash_t &h) {
    memcpy(hash,h.hash,16);
  }
};

#define PREAD
//#define ODIRECT

hash_t prehashfile(const string &s,hash_t &h1,ll_t size)
{
  static unsigned char prehash_buff[4096*3];
  static unsigned char *prehash_buff_=(prehash_buff+512)-((reinterpret_cast<ll_t>(prehash_buff+512))&511); // nice !
  
  int fd=open(s.c_str(),O_RDONLY
#ifdef ODIRECT
	      |O_DIRECT
#endif
	      );
  if(fd==-1) return hash_t(1);

  size_t r;
  unsigned char *buff=prehash_buff_;
  MD5_CTX ctx;
  if(!noend)
    MD5_Init(&ctx);

  MD5_CTX ctx1;
  MD5_Init(&ctx1);

  r=pread(fd,buff,4096,0);
  if(r!=4096) {
    fprintf(stderr,"error in read %s (prehash 0)\n",s.c_str());
    close(fd);
    h1.setx(1);
    return hash_t(1);
  }
  if(!noend)
    MD5_Update(&ctx,buff,r);
  MD5_Update(&ctx1,buff,r);

  if(!noend) {
    r=pread64(fd,buff,4096,4096*(size/8192));
    if(r!=4096) {
      fprintf(stderr,"error in read %s (prehash 1)\n",s.c_str());
      close(fd);
      return hash_t(1);
    }
    
    MD5_Update(&ctx,buff,r);

    r=pread(fd,buff,4096*2,4096*((size-2048)/4096));
    if(r!=size-4096*((size-2048)/4096)) {
      fprintf(stderr,"error in read %s (prehash 2)\n",s.c_str());
      close(fd);
      return hash_t(1);
    }
    
    MD5_Update(&ctx,buff,size-4096*((size-2048)/4096));
  }

  close(fd);
    
  hash_t h;
  MD5_Final(h1.hash,&ctx1);
  if(!noend)
    MD5_Final(h.hash,&ctx);
  else 
    h=h1;
  return h;
}

hash_t prehashfile(const string &s,hash_t &h1)
{
  int fd=open(s.c_str(),O_RDONLY);
  if(fd==-1) return hash_t(1);

  ll_t size=0;
  size_t r;
  if(!noend) {
    size=lseek64(fd,0,SEEK_END);
    if(size<0) {
      fprintf(stderr,"error in lseek %s (prehash)\n",s.c_str());
      close(fd);
      h1.setx(1);
      return hash_t(1);
    }
    r=lseek64(fd,0,SEEK_SET);
    if(r<0) {
      fprintf(stderr,"error in lseek %s (prehash)\n",s.c_str());
      close(fd);
      h1.setx(1);
      return hash_t(1);
    }
  }
 
  unsigned char buff[4096];
  MD5_CTX ctx;
  if(!noend)
    MD5_Init(&ctx);

  MD5_CTX ctx1;
  MD5_Init(&ctx1);

  r=read(fd,buff,4096);
  if(r!=4096) {
    fprintf(stderr,"error in read %s (prehash)\n",s.c_str());
    close(fd);
    h1.setx(1);
    return hash_t(1);
  }
  if(!noend)
    MD5_Update(&ctx,buff,r);
  MD5_Update(&ctx1,buff,r);

  if(!noend) {
    r=lseek64(fd,4096*(size/8192),SEEK_SET);
    if(r<0) {
      fprintf(stderr,"error in lseek %s (prehash)\n",s.c_str());
      close(fd);
      return hash_t(1);
    }
    r=read(fd,buff,4096);
    if(r!=4096) {
      fprintf(stderr,"error in read %s (prehash)\n",s.c_str());
      close(fd);
      return hash_t(1);
    }
    
    MD5_Update(&ctx,buff,r);
    
    r=lseek64(fd,-4096,SEEK_END);
    if(r<0) {
      fprintf(stderr,"error in lseek %s (prehash)\n",s.c_str());
      close(fd);
      return hash_t(1);
    }
    r=read(fd,buff,4096);
    if(r!=4096) {
      fprintf(stderr,"error in read %s (prehash)\n",s.c_str());
      close(fd);
      return hash_t(1);
    }
    
    MD5_Update(&ctx,buff,r);
  }

  close(fd);
    
  hash_t h;
  MD5_Final(h1.hash,&ctx1);
  if(!noend)
    MD5_Final(h.hash,&ctx);
  else 
    h=h1;
  return h;
}

#define HASH_CHK (8*1024*1024)
char hash_buff[HASH_CHK];

hash_t hashfile(const string &s)
{
  int fd=open(s.c_str(),O_RDONLY);
  if(fd==-1) return hash_t(1);
  
  MD5_CTX ctx;
  MD5_Init(&ctx);

  size_t r;
  do {
    r=read(fd,hash_buff,HASH_CHK);
    if(r==-1) {
      fprintf(stderr,"error in read %s\n",s.c_str());
      close(fd);
      return hash_t(1);
    }
    MD5_Update(&ctx,hash_buff,r);
  } while(r==HASH_CHK);

  close(fd);
    
  hash_t h;
  MD5_Final(h.hash,&ctx);
  return h;
}

hash_t hashbuff(const char *bf,int size)
{
  MD5_CTX ctx;
  MD5_Init(&ctx);
  MD5_Update(&ctx,bf,size);
  hash_t h;
  MD5_Final(h.hash,&ctx);
  return h;
}

struct tree_t {
  tree_t *p;
  tree_t *n;
  tree_t *e;
  tree_t *hl;
  string name;
  ll_t size;
  short type; // 0: file 1; dir, 2: other
  short flags;
  hash_t hash1;
  hash_t prehash;
  hash_t hash;
  dev_t dev;
  ino_t ino;

#define F_MOD 1
#define F_DEL 2

  void setmod() { flags|=F_MOD;}
  bool ismod() const { return flags&F_MOD;}
  void setdel() { flags|=F_DEL;setmod();}
  bool isdel() const { return flags&F_DEL;}
  bool ismod_rec() const {
    if(ismod()) return true;
    tree_t *p=e;
    while(p) {
      if(p->ismod()) return true;
      p=p->n;
    }
    return false;
  }

  tree_t() {
    hash.setx(1);
    prehash.setx(0);
    hash1.setx(0);
    p=n=e=NULL;
    size=-1;
    type=-1;
    hl=NULL;
    flags=0;
  }
  
  string getpath() const {
    if(p==NULL) return string(name);
    string r=p->getpath();
    return r+string("/")+string(name);
  }

  hash_t compprehash() {
    //md5 depuis fichier
    string p=getpath();
    hash_t h1;
#ifdef PREAD
    hash_t ph=prehashfile(p,h1,size);
#else
    hash_t ph=prehashfile(p,h1);
#endif
    prehash=ph;
    hash1=h1;
    return ph;
  }

  bool isparent (tree_t *b) const {
    if(b==NULL) return 0;
    if(this==b) return 1;
    return isparent(b->p);
  }

  hash_t comphash() {
    //md5 depuis fichier
    string p=getpath();
    hash=hashfile(p);
    return hash;
  }

  ll_t getsize() {
    string p=getpath();
    struct stat st;
    int r=lstat(p.c_str(),&st);
    if(r==-1) {
      printf("cannot stat %s\n",p.c_str());
      return -1;
    }
    size=st.st_size;
    return size;
  }

  ll_t compsizedir() {
    bool ok=1;
    ll_t r=0;
    tree_t *ptr=e;
    while(ptr) {
      if(ptr->size<0) ok=0;
      r+=ptr->size;
      ptr=ptr->n;
    }
    size=r;
    return size;
  }

  hash_t comphashdir() {
    //comp hash dir depuis hash des fils
    set<hash_t> sh;
    tree_t *ptr=e;
    while(ptr) {
      if(ptr->hash.isunk()) {
	hash.setunk();
	return hash;
      }
      if(ptr->hash.issol()) {
	hash.setsol();
	return hash;
      }
      if(!ptr->hash.isskip())
	sh.insert(ptr->hash);
      ptr=ptr->n;
    }

    if(sh.empty()) {
      hash.setskip();
      return hash;
    }
    
    char *bf=new char[16*sh.size()];
    int k=0;
    set<hash_t>::iterator it;
    for(it=sh.begin();it!=sh.end();++it) {
      memcpy(bf+16*k,it->hash,16);
      k++;
    }
    assert(k==sh.size());
    hash=hashbuff(bf,sh.size()*16);
    delete[]bf;
    return hash;
  }

  ll_t minsize() const {
    return size;
  }
};

void mds(ll_t &r,tree_t *a,set<hash_t> &s)
{
  tree_t *p=a->e;
  while(p) {
    if(p->type==0) {
      if(p->hash.issol()) r+=p->size;
      else if(!p->hash.isx()) {
	if(s.find(p->hash)==s.end()) {
	  s.insert(p->hash);
	  r+=p->size;
	}
      }
    }
    if(p->type==1)
      mds(r,p,s);
    p=p->n;
  }
}

ll_t mds(list<tree_t *>la)
{
  ll_t r=0;
  set<hash_t> s;
  list<tree_t *>::iterator it;
  for(it=la.begin();it!=la.end();++it)
    mds(r,*it,s);

  return r;
}

void ds(ll_t &r,tree_t *a,set<tree_t*> &s)
{
  tree_t *p=a->e,*p2;
  while(p) {
    if(p->type==0) {
      p2=p;
      while(p2->hl) p2=p2->hl;
      if(p2->hash.issol()) r+=p2->size;
      else if(!p2->hash.isx()) {
	if(s.find(p2)==s.end()) {
	  s.insert(p2);
	  r+=p2->size;
	}
      }
    }
    if(p->type==1)
      ds(r,p,s);
    p=p->n;
  }
}

ll_t ds(list<tree_t *>la)
{
  ll_t r=0;
  set<tree_t*> s;
  list<tree_t *>::iterator it;
  for(it=la.begin();it!=la.end();++it)
    ds(r,*it,s);

  return r;
}

ll_t nbrf=0;
ll_t nbrfail=0;
ll_t nbrd=0;
ll_t nbro=0;
tree_t *root;
map<ll_t,list<tree_t*> > mst;
map<ll_t,list<tree_t*> > mst_d;
map<hash_t,list<tree_t*> > mht;
map<hash_t,list<tree_t*> > mh1t;
map<hash_t,list<tree_t*> > mht_d;

void makemh1t(tree_t *t) {
  if(t->type==0 && !t->hash1.isx()) 
    mh1t[t->hash1].push_back(t);
  tree_t *p=t->e;
  while(p) {
    makemh1t(p);
    p=p->n;
  }
}

void maketree(const string &path,tree_t *tree) {
  //printf("%s\n",path.c_str());
  
  DIR *d=opendir(path.c_str());
  if(d==NULL) {
    fprintf(stderr,"cannot opendir %s\n",path.c_str());
    return;
  }
  tree_t *prev=tree;
  struct dirent *e;
  while(1) {
    e=readdir(d);
    if(e==NULL) break;
    if(0==strcmp(e->d_name,".")) continue;
    if(0==strcmp(e->d_name,"..")) continue;
    
    tree_t *ee=new tree_t();
    
    ee->p=tree;
    ee->name=e->d_name;
    if(prev==tree) tree->e=ee;
    else prev->n=ee;

    struct stat st;
    int r=lstat(ee->getpath().c_str(),&st);
    if(r==-1) {
      fprintf(stderr,"cannot stat %s\n",ee->getpath().c_str());
      ee->type=-1;
      ee->size=-1;
      nbrfail++;
    } else {
      ee->dev=st.st_dev;
      ee->ino=st.st_ino;

      if(S_ISDIR(st.st_mode)) {
	ee->type=1;
	nbrd++;
      }
      else if(S_ISREG(st.st_mode)) {
	nbrf++;
	ee->type=0;
	ee->size=st.st_size;
	if(ee->size>0)
	  mst[ee->size].push_back(ee);
	else
	  ee->hash.setskip();
      }
      else {
	nbro++;
	ee->type=2;
	ee->size=0;
	ee->hash.setskip();
      }
    }
    if((nbrfail+nbrf+nbrd+nbro)%10000==0) {
      fprintf(stderr,"%Ld files %Ld dirs %Ld others %Ld fails   \r",nbrf,nbrd,nbro,nbrfail);
    }

    prev=ee;
  }
  closedir(d);

  tree_t *ptr=tree->e;
  while(ptr) {
    if(ptr->type==1) {
      maketree(ptr->getpath(),ptr);
    }
    ptr=ptr->n;
  }

  ll_t s=tree->compsizedir();
  if(s>0)
    mst_d[s].push_back(tree);
}

tree_t *maketree(const char *path) {
  tree_t *root=new tree_t();
  root->type=1;
  root->name=string(path);
  maketree(string(path),root);
  return root;
}

void aff(tree_t *t) 
{
  if(t->hash.isunk())
    printf("%s: %Ld (%s) %d\n",t->getpath().c_str(),t->size,human(t->size),t->hash.isx()?int(t->hash.hash[0]):-1);
  tree_t *p=t->e;
  while(p) {
    aff(p);
    p=p->n;
  }
}

void makehashdir(tree_t *t) {
  if(t->type==0 && t->hl) {
    t->hash=t->hl->hash;
    return;
  }
  if(t->type==1) {
    tree_t *p=t->e;
    while(p) {
      makehashdir(p);
      p=p->n;
    }
    t->comphashdir();
    if(t->size>0 && !t->hash.isx()) {
      mht_d[t->hash].push_back(t);
    }
  }
} 

multimap<ll_t,tree_t*> mmlof;
void dirwithlotoffiles(tree_t *t) {
  if(t->type==1) {
    ll_t r=0;
    tree_t *p=t->e;
    while(p) {
      r++;
      dirwithlotoffiles(p);
      p=p->n;
    }
    if(r>100) mmlof.insert(make_pair(r,t));
  }
} 

bool isexactdir(tree_t *a,tree_t *b) 
{
  map<string,tree_t*> sb;
  tree_t *p,*p2;
  p=b->e;
  while(p) {
    sb[p->name]=p;
    p=p->n;
  }
  p=a->e;
  while(p) {
    if(sb.find(p->name)==sb.end()) return false;
    p2=sb[p->name];
    if(p->type!=p2->type 
       || (p->type==0 && 
	   (p->size!=p2->size || p->hash!=p2->hash || p->hash.issol()))) return false;
    p=p->n;
  }

  p=a->e;
  while(p) {
    p2=sb[p->name];
    if(p->type==1 && isexactdir(p,p2)==false) return false;
    p=p->n;
  }

  return true;
}

void dir(tree_t *a,map<hash_t,tree_t*> &s, ll_t &r)
{
  tree_t *p=a->e,*p2;
  while(p) {
    if(p->type==1)
      dir(p,s,r);

    if(p->type==0) {
      p2=p;
      while(p2->hl) p2=p2->hl;
      if(p2->hash.issol()) r+=p2->size;
      else if(!p2->hash.isx()) {
	if(s.find(p2->hash)==s.end()) {
	  s[p2->hash]=p2;
	}
      }
    }

    p=p->n;
  }
}

void cmpdir(tree_t *a,tree_t *b,ll_t &cap,ll_t &caph,ll_t &amb, ll_t &bma) 
{ 
  amb=bma=cap=caph=0;
  map<hash_t,tree_t*> sa,sb;
  dir(a,sa,amb);
  dir(b,sb,bma);
  map<hash_t,tree_t*>::iterator it;
  for(it=sa.begin();it!=sa.end();++it) {
    if(sb.find(it->first)==sb.end()) amb+=it->second->size;
    else if(it->second==sb.find(it->first)->second)
      caph+=it->second->size;
    else
      cap+=it->second->size;
  }
  for(it=sb.begin();it!=sb.end();++it)
    if(sa.find(it->first)==sa.end()) bma+=it->second->size;
}

struct cmpdir_t {
  ll_t amb,bma,cap,caph;
  bool disc;
};

set<pair<tree_t*,tree_t*> > scmpdir;
map<pair<tree_t*,tree_t*>,cmpdir_t> mcmpdir;

void cmpdirclean() {
  map<pair<tree_t*,tree_t*>,cmpdir_t>::iterator it,j;
  for(it=mcmpdir.begin();it!=mcmpdir.end();++it) {
    tree_t *a=it->first.first;
    tree_t *b=it->first.second;
    for(int i=1;i<4;i++) {
      tree_t *aa=a;
      tree_t *bb=b;
      if(i&1) aa=a->p;
      if(i&2) bb=b->p;
      if(aa==NULL ||bb==NULL) continue;
      if((j=mcmpdir.find(make_pair(aa,bb)))==mcmpdir.end()) continue;
      if(j->second.cap==it->second.cap &&j->second.caph==it->second.caph) j->second.disc=1;
    }
  }
}

void insertcmpdir(tree_t *a,tree_t*b) {
  if(a==NULL || b==NULL || a==b) return;
  if(a->isparent(b) ||b->isparent(a)) return;
  if(a>b) {insertcmpdir(b,a); return;}
  pair<tree_t*,tree_t*> p=make_pair(a,b);
  if(mcmpdir.find(p)!=mcmpdir.end()) return;
  scmpdir.insert(p);
}

double cmpdir_factor=0.1;

void proccmpdir(tree_t *a,tree_t *b)
{
  double f=cmpdir_factor;
  cmpdir_t c;
  c.amb=c.bma=c.cap=c.caph=0;
  c.disc=0;
  assert(a->type==1 && b->type==1);
  
  if(a->hash!=b->hash || a->hash.isx()) {
    cmpdir(a,b,c.cap,c.caph,c.amb,c.bma);
    mcmpdir[make_pair(a,b)]=c;
    ll_t i=c.cap+c.caph;
    ll_t m=c.amb;
    if(m>c.bma) m=c.bma;
    if(c.amb<f*i) insertcmpdir(a->p,b);
    if(c.bma<f*i) insertcmpdir(a,b->p);
    if(m<f*i) insertcmpdir(a->p,b->p);
  } else {
    c.cap=-1;
    mcmpdir[make_pair(a,b)]=c;
    insertcmpdir(a,b->p);
    insertcmpdir(a->p,b->p);
    insertcmpdir(a->p,b);
  }
}

void compcmpdir() 
{
  map<hash_t,list<tree_t*> >::iterator it2;
  
  for(it2=mht.begin();it2!=mht.end();++it2) {
    if(it2->second.size()>1) {
      list<tree_t *>::iterator i,j;
      for(i=it2->second.begin();i!=it2->second.end();++i) {
	for(j=it2->second.begin();j!=i;++j) {
	  insertcmpdir((*i)->p,(*j)->p);
	}
      }
    }
  }

  while(scmpdir.empty()==0) {
    pair<tree_t*,tree_t*> p=*(scmpdir.begin());
    scmpdir.erase(scmpdir.begin());

    proccmpdir(p.first,p.second);
  }
}


#if 0
//0:==, 1:< -1:>, 2:<>
int cmpdir(tree_t *a,tree_t *b,ll_t &cap,ll_t &amb,ll_t &bma) 
{
  amb=bma=cap=0;
  map<hash_t,tree_t*> sa,sb;
  map<string,tree_t*> sda,sdb;
  tree_t *p;
  p=a->e;
  while(p) {
    if(!(p->hash.isx()))
      sa[p->hash]=p;
    if(p->type==1) {
      sda[p->name]=p;
    }
    p=p->n;
  }

  p=b->e;
  while(p) {
    if(!(p->hash.isx()))
      sb[p->hash]=p;
    if(p->type==1) {
      sdb[p->name]=p;
    }
    p=p->n;
  }

  p=a->e;
  while(p) {
    if(p->type==0) {
      if(sb.find(p->hash)!=sb.end()) {
	cap+=p->size;
      } else {
	amb+=p->size;
      }
    }
    if(p->type==1) {
      if(sb.find(p->hash)!=sb.end()) {
	cap+=p->size;
      } else {
	amb+=p->size;
      }

    }
    p=p->n;
  }

  p=b->e;
  while(p) {
    if(sa.find(p->hash)!=sa.end()) {
      //cap+=sa[p->hash];
    } else {
      ssb[p->name]=p;
    }
    p=p->n;
  }
  
}
#endif

struct sit_t {
  dev_t dev;
  ino_t ino;
  tree_t *t;
};

int cmpsit(const void *aa,const void *bb)
{
  const sit_t *a=(const sit_t*)aa;
  const sit_t *b=(const sit_t*)bb;

  if(a->dev<b->dev) return -1;
  if(a->dev>b->dev) return 1;

  if(a->ino<b->ino) return -1;
  if(a->ino>b->ino) return 1;

  return 0;
}

int main(int ac, char **av)
{
  if(ac<=1) {
    printusage();
    return 1;
  }

  while(1) {
    if(0==strcmp(av[1],"-h")) {printusage(); return 0;}
    if(0==strcmp(av[1],"-H")) {printreadme(); return 0;}
    if(0==strcmp(av[1],"-n")) {md5=0;ac--;av++;continue;}
    if(0==strcmp(av[1],"-t")) {sortbysize=1;ac--;av++;continue;}
    if(0==strcmp(av[1],"-b")) {noend=1;ac--;av++;continue;}
    if(0==strcmp(av[1],"-r")) {minreportsize=atoll(av[2]);ac-=2;av+=2;continue;}
    if(0==strcmp(av[1],"-s")) {systematic=atoll(av[2]);ac-=2;av+=2;continue;}
    
    break;
  } 

  map<ll_t,list<tree_t*> >::iterator it;
  map<ll_t,list<tree_t*> >::reverse_iterator rit;

  ll_t ti;

  {
    root=new tree_t();
    tree_t *p=root;
    ll_t sr=0;
    for(int i=1;i<ac;i++) {
      ti=time(NULL);
      printf("scan %s\n",av[i]);
      tree_t *q=maketree(av[i]);
      q->type=1;
      if(p==root) root->e=q;
      else p->n=q;
      sr+=q->size;
      p=q;
      printf("time: %Ld minute(s)   \n",ll_t(time(NULL)-ti)/60);
    }
    root->size=sr;
    root->type=1;
  }

  printf("%Ld files %Ld dirs %Ld others %Ld fails   \n",nbrf,nbrd,nbro,nbrfail);
  printf("size: %Ld (%s)\n",root->minsize(),human(root->minsize()));

  printf("search hardlinks (hl.txt)...\n");

  {
    multimap<ll_t,pair<tree_t*,tree_t*> > mmhl;
    for(it=mst.begin();it!=mst.end();++it) {
      if(it->first>0 && it->second.size()>1) {
	map<pair<dev_t,ino_t>,list<tree_t*>::iterator > mhl;
	list<tree_t *>::iterator i,i2;
	for(i=it->second.begin();i!=it->second.end();) {
	  if(mhl.find(make_pair((*i)->dev,(*i)->ino))!=mhl.end()) {
	    mmhl.insert(make_pair(it->first,make_pair(*i,*(mhl.find(make_pair((*i)->dev,(*i)->ino))->second))));
	    
	    (*i)->hl=*mhl.find(make_pair((*i)->dev,(*i)->ino))->second;
	    
	    i2=i;++i2;
	    it->second.erase(i);
	    i=i2;
	  } else {
	    mhl[make_pair((*i)->dev,(*i)->ino)]=i;
	    ++i;
	  }
	}
      }
    }

    FILE *out=fopen("hl.txt","w");
    multimap<ll_t,pair<tree_t*,tree_t*> >::reverse_iterator itr;
    for(itr=mmhl.rbegin();itr!=mmhl.rend();++itr) {
      if(itr->first<minreportsize) break;
      fprintf(out,"  %s\n= %s\n",itr->second.first->getpath().c_str(),itr->second.second->getpath().c_str());
    }
    fclose(out);
  }

  //aff(root);
  //printf("hash: files...\n");
  ll_t bth=0;
  ll_t fth=0;
  for(it=mst.begin();it!=mst.end();++it) {
    if((it->first>0 && it->second.size()>1)
       || it->first>=systematic ) {
      fth+=it->second.size();
      bth+=it->second.size()*it->first;
    }
  }
  printf("to prehash: %Ld files\n",fth);

  ll_t bth_c=0;
  ll_t fth_c=0;
  int plog=1;
  ll_t ll=0;

  printf("sort files by dev/ino...\n");

  sit_t *sit=new sit_t[fth];
  ll_t k=0;

  for(rit=mst.rbegin();rit!=mst.rend();++rit) {
    map<hash_t,list<tree_t*> >::iterator pit;
    list<tree_t *>::iterator i;
    if((rit->first>0 && rit->second.size()>1) 
       || rit->first>=systematic ) {
      for(i=rit->second.begin();i!=rit->second.end();++i) {
	sit[k].dev=(*i)->dev;
	sit[k].ino=(*i)->ino;
	sit[k].t=(*i);
	k++;
      }
    } else  {
      for(i=rit->second.begin();i!=rit->second.end();++i) {
	if(rit->first==0) {
	  (*i)->hash.setskip();
	  (*i)->hash1.setskip();
	  (*i)->prehash.setskip();
	} 
	else {
	  (*i)->hash.setsol();
	  (*i)->hash1.setsol();
	  (*i)->prehash.setsol();
	}
      }
    }
  }

  assert(k==fth);
  printf("qsort...\n");
  qsort(sit,fth,sizeof(sit_t),cmpsit);
  
  printf("prehash...\n");
  ti=time(NULL);
  for(ll_t l=0;l<fth;l++) {
    if(l%1000==0) fprintf(stderr,"%Ld/%Ld  \r",l,fth);
    tree_t *t=sit[l].t;
    if(t->size<4096*5) {
      t->comphash();
      t->prehash=t->hash;
    } else {
      t->compprehash();
    }
  }
  printf("time: %Ld minutes   \n",ll_t(time(NULL)-ti)/60);
  delete[]sit;

  printf("hash...   \n");
  ti=time(NULL);
  for(rit=mst.rbegin();rit!=mst.rend();++rit) {
    map<hash_t,list<tree_t*> >::iterator pit;
    list<tree_t *>::iterator i;
    if(rit->first<10*1024*1024) {
      plog=0;
      if(ll<bth_c-10*1024*1024) {
	plog=0;
	char bf2[30];
	fprintf(stderr,"%Ld (%s) : %Ld files %Ld bytes (%s)  \r",rit->first,human(rit->first),fth_c,ll=bth_c,human(bth_c,bf2));
      }
    }

    if(rit->first>0 && rit->second.size()>1) {
      if(rit->first>=4096*5) {
	map<hash_t,list<tree_t*> > mpht;
	for(i=rit->second.begin();i!=rit->second.end();++i) {
	  hash_t ph=(*i)->prehash;
	  mpht[ph].push_back(*i);
	}
	for(pit=mpht.begin();pit!=mpht.end();++pit) {
	  if(pit->second.size()>1) {
	    for(i=pit->second.begin();i!=pit->second.end();++i) {
	      if(md5) 
		(*i)->comphash();
	      else {
		// computes a 'fake hash' with prehash and size
		(*i)->hash=(*i)->prehash;
		*(ll_t*)((*i)->hash.hash)=(*(ll_t*)((*i)->hash.hash))^((*i)->size);
	      }

	      fth_c+=1;
	      bth_c+=rit->first;
	      if(plog) {
		char bf2[10];
		fprintf(stderr,"%Ld (%s) : %Ld files %Ld bytes (%s)  \r",rit->first,human(rit->first),fth_c,bth_c,human(bth_c,bf2));
	      }

	      if(!(*i)->hash.isx()) {
		mht[(*i)->hash].push_back(*i);
	      }
	    }
	  } else {
	    for(i=pit->second.begin();i!=pit->second.end();++i) {
	      (*i)->hash.setsol();
	    }
	    fth_c+=1;
	    bth_c+=rit->first;
	    if(plog) {
	      char bf2[10];
	      fprintf(stderr,"%Ld (%s) : %Ld files %Ld bytes (%s)  \r",rit->first,human(rit->first),fth_c,bth_c,human(bth_c,bf2));
	    }
	  }
	}
      } else {
	for(i=rit->second.begin();i!=rit->second.end();++i) {

	  fth_c+=1;
	  bth_c+=rit->first;

	  if(!(*i)->hash.isx()) {
	    mht[(*i)->hash].push_back(*i);
	  }
	}
      }
    } else  {
      for(i=rit->second.begin();i!=rit->second.end();++i) {
	if(rit->first==0)
	  (*i)->hash.setskip();
	else
	  (*i)->hash.setsol();
      }
    }
  } 
  printf("time: %Ld minutes       \n",ll_t(time(NULL)-ti)/60);

  {
    map<hash_t,list<tree_t*> >::iterator it2;

    ll_t rec=0;
    for(it2=mht.begin();it2!=mht.end();++it2) {
      if(it2->second.size()>1) {
	rec+=(it2->second.size()-1)*(*it2->second.begin())->size;
      }
    }
    printf("max recuperable : %s\n", human(rec));

    printf("write id_files.txt\n");
    multimap<ll_t,list<tree_t*> > mmsh;
    for(it2=mht.begin();it2!=mht.end();++it2) {
      if(it2->second.size()>1) {
	list<tree_t*> l;
	list<tree_t *>::iterator i;
	ll_t s;
	for(i=it2->second.begin();i!=it2->second.end();++i) {
	  s=(*i)->size;
	  l.push_back(*i);
	}
	if(l.size()>1)
	  mmsh.insert(make_pair(s,l));
      }
    }
    
    FILE *out=fopen("id_files.txt","w");
    multimap<ll_t,list<tree_t*> >::reverse_iterator itr;
    for(itr=mmsh.rbegin();itr!=mmsh.rend();++itr) {
      if(itr->first<minreportsize) break;
      list<tree_t*>::iterator i;
      for(i=itr->second.begin();i!=itr->second.end();++i) {
	fprintf(out,"%s ( %Ld = %s )\n",(*i)->getpath().c_str(),(*i)->size,human((*i)->size));
      }
      fprintf(out,"====================\n");
    }
    fclose(out);
  }

  makemh1t(root);

  {
    map<hash_t,list<tree_t*> >::iterator it2;
    
    printf("write sim_files.txt %Ld\n",ll_t(mh1t.size()));
    multimap<ll_t,map<hash_t,list<tree_t*> > > mmsh;
    for(it2=mh1t.begin();it2!=mh1t.end();++it2) {
      if(it2->second.size()>1) {
	list<tree_t *>::iterator i;
	map<hash_t,list<tree_t*> > ml;
	ll_t s=-1,s2=-1;
	int k=1;
	for(i=it2->second.begin();i!=it2->second.end();++i) {
	  if((*i)->hash.issol()) {
	    hash_t h=(*i)->hash;
	    *(int*)h.hash=(*(int*)h.hash)^k; // fake hash for 'sol' files
	    k++;
	    ml[h].push_back(*i);
	  } else {
	    if(!(*i)->hash.isx())
	      ml[(*i)->hash].push_back(*i);
	  }
	  if(s==-1 || (*i)->size>s) {s2=s;s=(*i)->size;}
	  else
	    if(s2==-1 || (*i)->size>s2) {s2=(*i)->size;}
	}
	if(ml.size()>1) {
	  assert(s2>0);
	  mmsh.insert(make_pair(s2,ml));
	}
      }
    }
    
    FILE *out=fopen("sim_files.txt","w");
    multimap<ll_t,map<hash_t,list<tree_t*> > >::reverse_iterator itr;
    for(itr=mmsh.rbegin();itr!=mmsh.rend();++itr) {
      if(itr->first<minreportsize) break;
      map<hash_t,list<tree_t*> >::iterator i;
      fprintf(out,"=== size = %Ld = %s ===\n",itr->first,human(itr->first));
      for(i=itr->second.begin();i!=itr->second.end();++i) {
	if(i!=itr->second.begin()) 
	  fprintf(out,"------\n");
	list<tree_t*>::iterator j;
	for(j=i->second.begin();j!=i->second.end();++j) {
	  fprintf(out,"%s ( %Ld = %s )\n",(*j)->getpath().c_str(),(*j)->size,human((*j)->size));
	}
      }
      fprintf(out,"====================\n");
    }
    fclose(out);
  }

  ti=time(NULL);
  printf("hash: dirs...\n");
  makehashdir(root);
  printf("time: %Ld minutes\n",ll_t(time(NULL)-ti)/60);

  //printf("%s (%Ld)\n",(*i)->getpath().c_str(),(*i)->size);

  printf("write id_dirs.txt\n");
  {
    map<hash_t,list<tree_t*> >::iterator it2;
    multimap<ll_t,list<tree_t*> > mmsh;
    for(it2=mht_d.begin();it2!=mht_d.end();++it2) {
      if(it2->second.size()>1) {
	list<tree_t*> l;
	list<tree_t *>::iterator i;
	ll_t s;
	for(i=it2->second.begin();i!=it2->second.end();++i) {
	  s=(*i)->size;
	  l.push_back(*i);
	}
	if(!sortbysize) s=ds(l)-mds(l);
	if(l.size()>1)
	  mmsh.insert(make_pair(s,l));
      }
    }
    
    FILE *out=fopen("id_dirs.txt","w");
    multimap<ll_t,list<tree_t*> >::reverse_iterator itr;
    for(itr=mmsh.rbegin();itr!=mmsh.rend();++itr) {
      if(itr->first<minreportsize) break;
      fprintf(out,"=== save %s ===\n",human(ds(itr->second)-mds(itr->second)));
      list<tree_t*>::iterator i;
      for(i=itr->second.begin();i!=itr->second.end();++i) {
	fprintf(out,"%s ( %s%s %Ld = %s )\n",(*i)->getpath().c_str(),i==itr->second.begin()||isexactdir(*(itr->second.begin()),*i)?"":"NO-",i==itr->second.begin()?"":"EXACT",(*i)->size,human((*i)->size));
      }
    }
    fclose(out);
  }

  {
    FILE *out=fopen("lof_dirs.txt","w");
    printf("search for dirs with a lot of files (lof_dirs.txt)\n");
    dirwithlotoffiles(root);
    multimap<ll_t,tree_t*>::reverse_iterator itr;
    for(itr=mmlof.rbegin();itr!=mmlof.rend();++itr) {
      fprintf(out,"%s ( %Ld files )\n",itr->second->getpath().c_str(),itr->first);
    }
    fclose(out);
  }

  aff(root);

  {
    ti=time(NULL);
    printf("search similars dirs...\n");
    compcmpdir();
    cmpdirclean();
    printf("time: %Ld minutes\n",ll_t(time(NULL)-ti)/60);

  /*
    map<pair<tree_t*,tree_t*>,cmpdir_t>::iterator it;
    for(it=mcmpdir.begin();it!=mcmpdir.end();++it) {
      tree_t *a=it->first.first;
      tree_t *b=it->first.second;
      cmpdir_t c=it->second;
      printf("%s %s %Ld %Ld %Ld %Ld\n",a->getpath().c_str(),b->getpath().c_str(),c.cap,c.caph,c.amb,c.bma);
      //if(a->hash.isx()) printf("a: %d\n",int(a->hash.hash[0]));
      //if(b->hash.isx()) printf("b: %d\n",int(b->hash.hash[0]));
    }
  */

    map<pair<tree_t*,tree_t*>,cmpdir_t>::iterator it2;
    multimap<ll_t,pair<pair<tree_t*,tree_t*>,cmpdir_t> > mmsh;
    for(it2=mcmpdir.begin();it2!=mcmpdir.end();++it2) {
      if(it2->second.cap==-1) continue;
      if(it2->second.disc) continue;
      ll_t s=it2->second.cap;
      if(!sortbysize) s+=it2->second.caph;
      mmsh.insert(make_pair(s,make_pair(it2->first,it2->second)));
    }
    
    FILE *out=fopen("sim_dirs.txt","w");
    multimap<ll_t,pair<pair<tree_t*,tree_t*>,cmpdir_t> >::reverse_iterator itr;
    for(itr=mmsh.rbegin();itr!=mmsh.rend();++itr) {
      if(itr->first<minreportsize) break;
      tree_t *a=itr->second.first.first;
      tree_t *b=itr->second.first.second;
      cmpdir_t c=itr->second.second;
      fprintf(out,"=== save %s ===\n",human(itr->second.second.cap));
      fprintf(out," %Ld (hl: %Ld) < %Ld  > %Ld\n",c.cap,c.caph,c.amb,c.bma);
      fprintf(out,"%s\n%s\n",a->getpath().c_str(),b->getpath().c_str());
    }
    fclose(out);

  }

  return 0;
}
